import numpy as np


class SKT:
    """Sheguesada-Kawazaki-Teramoto population interaction model.

    Attributes:
        D (np.ndarray): diffusion matrix.
        R (np.ndarray): reaction matrix.
    """

    def __init__(self, D: np.ndarray, R: np.ndarray) -> None:
        self.D, self.R = D, R


class LV(SKT):
    """Lotka-Volterra population interaction model. Coefficients must all be positive.

    Attributes:
        a (float): prey repoduction rate.
        b (float): prey mortality due to predators.
        c (float): predator reproduction due to eating preys.
        d (float): predator death rate.
    """

    def __init__(self, a: float, b: float, c: float, d: float) -> None:
        if not (a >= 0 and b >= 0 and c >= 0 and d >= 0):
            raise ValueError("Coefficients must all be positive")
        self.D = np.zeros((2, 3))
        self.R = np.array([[a, 0, b], [-d, -c, 0]])


class CLV(SKT):
    """Competitive Lotka-Volterra population interaction model. Coefficients must all be positive.

    See https://en.wikipedia.org/wiki/Competitive_Lotka-Volterra_equations.

    Attributes:
        r1 (float): reproduction rate for (1).
        r2 (float): reproduction rate for (2).
        K1 (float): carrying capacity for (1).
        K2 (float): carrying capacity for (2).
        s12 (float): competitive effect (1) has on (2).
        s21 (float): competitive effect (2) has on (1).
    """

    def __init__(
        self, r1: float, r2: float, K1: float, K2: float, s12: float, s21: float
    ) -> None:
        if not (r1 >= 0 and r2 >= 0 and s12 >= 0 and s21 >= 0):
            raise ValueError("Coefficients must all be positive")
        if not (K1 > 0 and K2 > 0):
            raise ValueError("Carrying capacities must be > 0")
        self.D = np.zeros((2, 3))
        self.R = np.array(
            [[r1, -r1 / K1, -r1 * s12 / K1], [r2, -r2 * s21 / K2, -r2 / K2]]
        )



import numpy as np


def XtoUV(X: np.ndarray) -> tuple[np.ndarray, np.ndarray]:
    """Recovers the concentration vector of species U and V from the concatenated vector.

    Args:
        X (np.ndarray): Concatenated vector.

    Returns:
        tuple[np.ndarray, np.ndarray]: Species concentration vectors.
    """
    K = len(X) // 2
    return X[:K], X[K:]


def UVtoX(U: np.ndarray, V: np.ndarray) -> np.ndarray:
    """Concatenates the concentration vector of species U and V.

    Args:
        U (np.ndarray): First species concentration vector.
        V (np.ndarray): Second species concentration vector.

    Returns:
        np.ndarray: Concatenated vectors.
    """
    K = len(U)
    X = np.zeros(2 * K)
    X[:K] = U
    X[K:] = V
    return X


def merge_diags(
    P: list[np.ndarray], Q: list[np.ndarray], R: list[np.ndarray], S: list[np.ndarray]
) -> list[np.ndarray]:
    """Assemble the list of diagonals of the Jacobian from the block data.

    Args:
        P (list[np.ndarray]): upper left Jacobian block.
        Q (list[np.ndarray]): upper right Jacobian block.
        R (list[np.ndarray]): lower left Jacobian block.
        S (list[np.ndarray]): lower right Jacobian block.

    Returns:
        list[np.ndarray]: List of diagonals for full Jacobian.
    """
    return [
        Q[0],
        Q[1],
        Q[2],
        np.concatenate((P[0], Q[3], S[0])),
        np.concatenate((P[1], Q[4], S[1])),
        np.concatenate((P[2], S[2])),
        np.concatenate((P[3], R[0], S[3])),
        np.concatenate((P[4], R[1], S[4])),
        R[2],
        R[3],
        R[4],
    ]


def f(i: int, A: np.ndarray, B: np.ndarray, D: np.ndarray, R: np.ndarray) -> np.ndarray:
    """Helper function involved in the calculation of the functional to be cancelled. See the theory for more details.

    Args:
        i (int): species index (0 or 1).
        A (np.ndarray): first concentration vector.
        B (np.ndarray): second concentration vector.
        D (np.ndarray): re-normalized diffusion matrix.
        R (np.ndarray): re-normalized reaction matrix.

    Returns:
        np.ndarray: Output vector.
    """
    assert i == 0 or i == 1
    A2, AB = A * A, A * B
    center = (
        ((1 - R[i, 0] + 2 * D[i, 0]) * A)
        + ((R[i, i + 1] + 2 * D[i, i + 1]) * A2)
        + ((R[i, 2 - i] + 2 * D[i, 2 - i]) * AB)
    )
    border = D[i, 0] * A + D[i, i + 1] * A2 + D[i, 2 - i] * AB
    return center - np.roll(border, 1) - np.roll(border, -1)


def mu(i: int, A: np.ndarray, B: np.ndarray, D: np.ndarray) -> np.ndarray:
    """Helper function involved in the calculation of the Jacobian. See the theory for more details.

    Args:
        i (int): species index.
        A (np.ndarray): first concentration vector.
        B (np.ndarray): second concentration vector.
        D (np.ndarray): re-normalized diffusion matrix.

    Returns:
        np.ndarray: Output vector.
    """
    assert i == 0 or i == 1
    return D[i, 0] * np.ones(len(A)) + 2 * D[i, i + 1] * A + D[i, 2 - i] * B


def nu(
    i: int, A: np.ndarray, B: np.ndarray, D: np.ndarray, R: np.ndarray
) -> np.ndarray:
    """Helper function involved in the calculation of the Jacobian. See the theory for more details.

    Args:
        i (int): species index.
        A (np.ndarray): first concentration vector.
        B (np.ndarray): second concentration vector.
        D (np.ndarray): re-normalized diffusion matrix.

    Returns:
        np.ndarray: Output vector.
    """
    assert i == 0 or i == 1
    return (
        (1 - R[i, 0] + 2 * D[i, 0]) * np.ones(len(A))
        + 2 * (R[i, i + 1] + 2 * D[i, i + 1]) * A
        + (R[i, 2 - i] + 2 * D[i, 2 - i]) * B
    )


def block_diags(
    i: int, A: np.ndarray, D: np.ndarray, R: np.ndarray
) -> list[np.ndarray]:
    """Used to compute two of the Jacobian's blocks. See the theory for more details.

    Args:
        i (int): species index (0 or 1).
        A (np.ndarray): first concentration vector.
        D (np.ndarray): re-normalized diffusion matrix.
        R (np.ndarray): re-normalized reaction matrix.

    Returns:
        list[np.ndarray]: List of diagonals for the specified block.
    """
    assert i == 0 or i == 1
    DA = D[i, 2 - i] * A
    return [
        -DA[-1:],
        -DA[1:],
        R[i, 2 - i] * A + 2 * DA,
        -DA[:-1],
        -DA[:1],
    ]



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="PySpecies",
    version="0.1.2",
    author="Samuel Boïté, Mathias Grau",
    author_email="boite.samuel@gmail.com",
    description="Blazing-fast simulation of self-organized patterns in reaction-diffusion systems.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/samuelbx/pyspecies",
    package_dir={"": "src"},
    packages=setuptools.find_packages(where="src"),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.6",
)



import warnings
from math import ceil
from typing import Callable

import matplotlib.pyplot as plt
import numpy as np
from matplotlib.animation import FuncAnimation

from pyspecies._euler import back_euler
from pyspecies._utils import UVtoX, XtoUV
from pyspecies.models import SKT


class Pop:
    """A class representing two populations and their interaction model.

    Parameters:
        u0 (Callable): initial concentration of species #1
        v0 (Callable): initial concentration of species #2
        model (SKT): species interaction model
        space (tuple[int, int, int], optional): tuple containing (min_x, max_x, no_points). Defaults to (0, 1, 200).
    """

    def __init__(
        self,
        u0: Callable,
        v0: Callable,
        model: SKT,
        space: tuple[int, int, int] = (0, 1, 200),
    ) -> None:
        if len(space) != 3:
            raise ValueError("space must contain 3 elements: min_x, max_x, no_points")
        if space[0] > space[1]:
            raise ValueError("max_x must be greater than min_x")
        K = space[2]
        if K < 10:
            warnings.warn(
                "The space must have at least 10 points for the program to work properly. The number of points has been automatically set to 10."
            )
            K = 10

        # Store diffusion and reaction matrices and discretized space
        self.D, self.R = model.D, model.R
        self.Space = np.linspace(space[0], space[1], K)

        # Compute concentration vectors of initial population
        X0 = UVtoX(u0(self.Space), v0(self.Space))
        if not (X0 >= 0).all():
            raise ValueError("Initial conditions must be positive")

        # Initialize simulation results
        self.Xlist = [X0]
        self.Tlist = np.array([0])

    def sim(self, duration: float, N: int = 100) -> None:
        """Move the simulation forward by a given duration and precision.

        Args:
            duration (float): duration of the simulation in seconds
            N (int, optional): number of time steps (defaults to 100)
        """
        if duration <= 0:
            raise ValueError("Duration must be positive")
        if N < 1:
            raise ValueError("N must be greater than one")

        # Generate corresponding time steps
        Time = np.linspace(0, duration, N)

        # Continue the simulation and save the results
        X0 = self.Xlist[-1].copy()
        self.Xlist += back_euler(X0, Time, self.Space, self.D, self.R)
        self.Tlist = np.append(self.Tlist, self.Tlist[-1] + Time)

    def _formatPlot(self):
        """Helper function to prepare the concentration vs. space graph."""
        # Chart format
        plt.style.use("seaborn-talk")
        fig, ax = plt.subplots()
        ax.set_xlabel("Space")
        ax.set_ylabel("Concentrations")

        # Define the size of the graph axis
        xmin, xmax = np.min(self.Space), np.max(self.Space)
        padx = (xmax - xmin) * 0.05
        ymin, ymax = np.min(self.Xlist), np.max(self.Xlist)
        pady = (ymax - ymin) * 0.05
        ax.set_xlim(xmin - padx, xmax + padx)
        ax.set_ylim(ymin - pady, ymax + 2 * pady)

        # Add a text for the simulation time
        time_text = ax.text(
            0.5,
            0.95,
            "",
            horizontalalignment="center",
            verticalalignment="center",
            transform=ax.transAxes,
        )

        return fig, ax, time_text

    def anim(self, length: float = 7) -> None:
        """Shows an elegant Matplotlib animation of the steps simulated so far.

        Args:
            length (int, optional): In theory, the duration of the animation in seconds. In practice, it will be relatively longer because of the time needed to render the different images. Defaults to 7.
        """
        if length <= 0:
            raise ValueError("Length must be positive")

        # Prepare the plot
        fig, ax, time_text = self._formatPlot()
        tmax = np.max(self.Tlist)

        # Function called to update the frames
        def _anim(i):
            # Find the closest time step
            j = ceil(i * (len(self.Xlist) - 1) / (length * 50 - 1))

            # Update the graph
            U, V = XtoUV(self.Xlist[j])
            Uarea = ax.fill_between(
                self.Space, U, color="#f44336", alpha=0.5, label="First species"
            )
            Varea = ax.fill_between(
                self.Space, V, color="#3f51b5", alpha=0.5, label="Second species"
            )

            # Update the time text
            display_time = np.round(self.Tlist[j], decimals=2)
            display_percentage = np.round(100 * self.Tlist[j] / tmax, decimals=1)
            time_text.set_text(
                f"Simulation at t={display_time}s ({display_percentage}%)"
            )

            return Uarea, Varea, time_text

        # Animate the graph
        ani = FuncAnimation(
            fig, _anim, frames=range(length * 50), interval=20, blit=True, repeat=True
        )
        plt.legend()
        plt.show()

    def snapshot(self, theta: float, show_prod: False) -> None:
        """Shows a snapshot of the state of the simulation at a certain percentage of completion.

        Args:
            theta (float): simulation completion level, between 0 (initial) and 1 (final)
            show_prod (bool, optional): whether to show the product of the two species. Defaults to False.
        """
        if not (0 <= theta and theta <= 1):
            raise ValueError("theta must lie between 0 and 1")

        # Prepare the plot
        _, ax, time_text = self._formatPlot()
        tmax = np.max(self.Tlist)

        # Find the closest time step
        j = 0
        while self.Tlist[j] < theta * tmax and j < len(self.Tlist) - 2:
            j += 1

        # Update the graph
        U, V = XtoUV(self.Xlist[min(j, len(self.Xlist) - 1)])
        ax.fill_between(
            self.Space, U, color="#f44336", alpha=0.5, label="First species"
        )
        ax.fill_between(
            self.Space, V, color="#3f51b5", alpha=0.5, label="Second species"
        )
        if show_prod:
            ax.plot(self.Space, U * V, color="red")

        # Update text for simulation time
        display_time = np.round(self.Tlist[j], decimals=2)
        time_text.set_text(f"Simulation at t={display_time}s ({theta * 100:.0f}%)")
        plt.legend()
        plt.show()

    def heatmap(self) -> None:
        """Displays a nice 2D heatmap of dominant species over time and space."""
        # Prepare the data
        grid = np.zeros((len(self.Tlist), len(self.Space)))
        for i, X in enumerate(self.Xlist):
            U, V = XtoUV(X)
            grid[i, :] = U - V
        grid = grid[:-1, :-1]

        # Show the heatmap
        fig, ax = plt.subplots()
        Ubound, Vbound = max(grid.max(), 0), max(-grid.min(), 0)
        bound = max(Ubound, Vbound)  # Color palette boundary
        c = ax.pcolormesh(
            self.Space, self.Tlist, grid, cmap="RdBu_r", vmin=-bound, vmax=bound
        )
        ax.set_title("Domination heatmap (u: red, v: blue)")
        ax.set_xlabel("Space")
        ax.set_ylabel("Time")
        fig.colorbar(c, ax=ax)
        plt.show()



import numpy as np
import scipy.sparse as sp
from scipy.sparse.linalg import spsolve
from tqdm import tqdm

from pyspecies._utils import XtoUV, block_diags, f, merge_diags, mu, nu


def _funcjac(
    X: np.ndarray, Xm: np.ndarray, D: np.ndarray, R: np.ndarray, dx: float, dt: float
) -> tuple[np.ndarray, sp.dia_matrix]:
    """Computes the function and the sparse Jacobian used in the iteration of Newton's method.

    Args:
        X (np.ndarray): current concentration vector.
        Xm (np.ndarray): previous concentration vector.
        D (np.ndarray): diffusion matrix.
        R (np.ndarray): reaction matrix.
        dx (float): space step.
        dt (float): time step.

    Returns:
        tuple[np.ndarray, sp.dia_matrix]: function and Jacobian.
    """

    # Current and previous concentration vectors
    U, V = XtoUV(X)
    Um, Vm = XtoUV(Xm)

    D2, R2 = dt / dx**2 * D, dt * R  # Renormalize matrices
    K = len(U)  # Number of space points

    # Compute function
    g = np.concatenate((f(0, U, V, D2, R2) - Um, f(1, V, U, D2, R2) - Vm))

    # Compute jacobian blocks
    r, s = mu(0, U, V, D2), mu(1, V, U, D2)
    JP = [-r[-1:], -r[1:], nu(0, U, V, D2, R2), -r[:-1], -r[:1]]
    JQ, JR = block_diags(0, U, D2, R2), block_diags(1, V, D2, R2)
    JS = [-s[-1:], -s[1:], nu(1, V, U, D2, R2), -s[:-1], -s[:1]]

    # Assemble jacobian matrix in sparse format
    jac = sp.diags(
        merge_diags(JP, JQ, JR, JS),
        [2 * K - 1, K + 1, K, K - 1, 1, 0, -1, -K + 1, -K, -K - 1, -2 * K + 1],
        format="csr",
    )

    return g, jac


def _cuthill_mckee(K: int) -> np.ndarray:
    """Cuthill-McKee algorithm for bandwidth reduction.

    Args:
        K (int): number of space points.

    Returns:
        np.ndarray: reverse Cuthill-McKee permutation.
    """

    # Assemble typical Jacobian matrix
    a = np.ones(K)
    d = [a[:1], a[1:], a, a[1:], a[:1]]
    M = sp.diags(
        merge_diags(d, d, d, d),
        [2 * K - 1, K + 1, K, K - 1, 1, 0, -1, -K + 1, -K, -K - 1, -2 * K + 1],
        format="csr",
    )

    # Compute reverse Cuthill-McKee permutation
    return sp.csgraph.reverse_cuthill_mckee(M, symmetric_mode=True)


def back_euler(
    X0: np.ndarray,
    Time: np.ndarray,
    Space: np.ndarray,
    D: np.ndarray,
    R: np.ndarray,
    newt_thres: float = 1e-4,
    max_iter: int = 10,
) -> list[np.ndarray]:
    """Solves the reaction-diffusion system using the backward Euler method.

    Args:
        X0 (np.ndarray): initial concentration vector.
        Time (np.ndarray): time steps.
        Space (np.ndarray): space steps.
        D (np.ndarray): diffusion matrix.
        R (np.ndarray): reaction matrix.
        newt_thres (float, optional): convergence threshold for Newton's method (defaults to 1e-4)
        max_iter (int, optional): maximum number of iterations for Newton's method (defaults to 10)

    Returns:
        list[np.ndarray]: approximate solution at each time step.
    """
    X_list = [X0]
    dx = Space[1] - Space[0]

    # Compute reverse Cuthill-McKee permutation
    perm = _cuthill_mckee(len(X0) // 2)

    for n in tqdm(range(1, len(Time)), "Deterministic simulation"):
        Xm = X_list[-1].copy()
        Xk = Xm.copy()
        dt = Time[n] - Time[n - 1]

        # Newton's method with sparse Jacobian
        for i in range(max_iter):
            # Compute function and Jacobian
            b, A = _funcjac(Xk, Xm, D, R, dx, dt)

            # Solve linear system after applying permutation
            deltaX = spsolve(A[perm, :][:, perm], -b[perm])[np.argsort(perm)]
            Xk += deltaX

            # Stop if convergence is reached
            if np.linalg.norm(deltaX) < newt_thres * np.linalg.norm(Xk):
                break

            # Raise error if maximum number of iterations is reached
            if i == max_iter - 1:
                raise ValueError(
                    f"Newton's method cannot converge in less than {max_iter} steps: aborting."
                )

        X_list.append(Xk)

    # Remove initial condition as it is the final condition of the previous simulation
    del X_list[0]

    return X_list
